from threading import Thread

from .usb_camera import USBCamera


class AsyncUSBCamera(USBCamera):
    """
    a usb camera class that reads from the device on another thread
    use this when multiple threads are trying to read from the camera at once
    """

    def __init__(self, port, data):
        USBCamera.__init__(self, port, data)
        self._ok, self._frame = False, None
        self._thread = Thread(target=self._async_camera_read)
        self._thread.start()

    def read(self, image=None):
        return self._ok, self._frame

    def _async_camera_read(self):
        while self.is_opened():
            self._ok, self._frame = USBCamera.read(self)

    def open(self, filename):
        self.release()
        USBCamera.open(self, filename)
        self._thread = Thread(target=self._async_camera_read)
        self._thread.start()



from .camera import Camera
from .stream_camera import StreamCamera
from .camera_data import CameraData
from .usb_camera import USBCamera
from .usb_stream_camera import USBStreamCamera
from .threshold import Threshold
from .threshold_group import ThresholdGroup
from .pipeline import PipeLine
from .game_object import GameObject
from .async_usb_camera import AsyncUSBCamera
from .camera_list import CameraList


import cv2
import numpy as np

from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.models.contours import find_contours, contours_to_circles, contours_to_rects, contours_to_rotated_rects,\
    contours_to_ellipses
from gbvision.utils.pipeline import PipeLine


class DrawContours(PipeLine):
    def __init__(self, threshold_func, color, thickness=2, contours_process=EMPTY_PIPELINE):
        contour_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process

        def _draw(frame):
            frame = frame.copy()
            cnts = contour_finding(frame)
            cv2.drawContours(frame, cnts, -1, color, thickness)
            return frame

        PipeLine.__init__(self, _draw)


class DrawCircles(PipeLine):
    def __init__(self, threshold_func, color, thickness=2, contours_process=EMPTY_PIPELINE,
                circle_process=EMPTY_PIPELINE):
        circle_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_circles + \
                         circle_process

        def _draw(frame):
            frame = frame.copy()
            circs = circle_finding(frame)
            for c in circs:
                cv2.circle(frame, (int(c[0][0]), int(c[0][1])), int(c[1]), color, thickness)
            return frame

        PipeLine.__init__(self, _draw)


class DrawRects(PipeLine):
    def __init__(self, threshold_func, color, thickness=2, contours_process=EMPTY_PIPELINE,
              rects_process=EMPTY_PIPELINE):
        rect_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_rects + \
                       rects_process

        def _draw(frame):
            frame = frame.copy()
            rects = rect_finding(frame)
            for r in rects:
                cv2.rectangle(frame, (int(r[0]), int(r[1])), (int(r[0] + r[2]), int(r[1] + r[3])), color, thickness)
            return frame

        PipeLine.__init__(self, _draw)


class DrawRotatedRects(PipeLine):
    def __init__(self, threshold_func, color, thickness=2, contours_process=EMPTY_PIPELINE,
                     rotated_rects_process=EMPTY_PIPELINE):
        rotated_rect_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_rotated_rects + \
                               rotated_rects_process

        def _draw(frame):
            frame = frame.copy()
            rotated_rects = rotated_rect_finding(frame)
            for r in rotated_rects:
                box = cv2.boxPoints(r)
                box = np.int0(box)
                cv2.drawContours(frame, [box], 0, color, thickness)
            return frame

        PipeLine.__init__(self, _draw)


class DrawEllipses(PipeLine):
    def __init__(self, threshold_func, color, thickness=2, contours_process=EMPTY_PIPELINE,
                 ellipses_process=EMPTY_PIPELINE):
        ellipses_finding = EMPTY_PIPELINE + threshold_func + find_contours + contours_process + contours_to_ellipses + \
                               ellipses_process

        def _draw(frame):
            frame = frame.copy()
            ellipses = ellipses_finding(frame)
            for e in ellipses:
                cv2.ellipse(frame, e, color, thickness)
            return frame

        PipeLine.__init__(self, _draw)



from os.path import splitext
from threading import Thread

import cv2

from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.constants.video import VIDEO_FILE_TYPE
from .feed_window import FeedWindow


class RecordingFeedWindow(FeedWindow):
    """
    a basic window that displays the stream from a stream receiver
    """

    def __init__(self, file_name: str, window_name: str, fps=20.0, exit_button='qQ',
                 drawing_pipeline=EMPTY_PIPELINE, recording_pipeline=EMPTY_PIPELINE):
        """
        initializes the feed window
        :param file_name: the name of the output file
        :param drawing_pipeline: optional, a pipeline of drawing functions that will run on the frame before displaying
        it
        :param exit_button: an array of keys (a string), when one of the keys are pressed the window will be closed
        :param window_name: the title of the window
        :param fps: the fps of the video file
        :param recording_pipeline: optional, a drawing pipeline to run on the frames being recorded
        """
        FeedWindow.__init__(self, window_name=window_name, exit_button=exit_button, drawing_pipeline=drawing_pipeline)
        self.recording_pipeline = recording_pipeline
        self.file_name = file_name

        _, file_ext = splitext(file_name)

        self.fourcc = cv2.VideoWriter_fourcc(*VIDEO_FILE_TYPE[file_ext.upper()])

        self.fps = fps
        self.video_writer = None

    def show_frame(self, frame):
        """
        display the stream video window
        :param frame: the frame to show
        """
        if self.video_writer is None and frame is not None:
            self.video_writer = cv2.VideoWriter(self.file_name, self.fourcc, self.fps, frame.shape[:2][::-1])
        self.video_writer.write(self.recording_pipeline(frame))
        return FeedWindow.show_frame(self, frame)



from .basic_ops import corners, edges, sharpen, blur, blue, red, green, gray
from .contours import contour_center, contours_centers, contours_to_circles, contours_to_circles_sorted, \
    contours_to_ellipses, contours_to_ellipses_sorted, contours_to_polygons, contours_to_rects, \
    contours_to_rects_sorted, contours_to_rotated_rects, contours_to_rotated_rects_sorted, FilterContours, \
    find_contours, sort_contours, sort_ellipses, sort_rotated_rects, sort_rects, sort_circles
from .shapes import circle_collision, filter_inner_circles, filter_inner_rects, rect_collision, rotated_rect_collision, \
    filter_inner_rotated_rects, convex_shape_collision, filter_inner_convex_shapes
from .denoising import Dilate, Erode, MedianBlur
